@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2004-2008  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2004-2009  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -25,13 +25,11 @@
  *  SUCH DAMAGE.
  *
  *
- *  $Id: debugger_expr.c,v 1.12.2.1 2008/01/18 19:12:27 debug Exp $
- *
  *  Expression evaluator.
  *
  *
  *  TODO:
- *	Sign-extension only on MIPS?
+ *	Sign-extension is arch dependent (e.g. MIPS has it).
  *	SPECIAL IMPORTANT CASE: Clear the delay_slot flag when writing
  *		to the pc register.
  *	TAB completion? :-)
