@@ -2,7 +2,7 @@
 #define	CPU_H
 
 /*
- *  Copyright (C) 2005-2008  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -28,8 +28,6 @@
  *  SUCH DAMAGE.
  *
  *
- *  $Id: cpu.h,v 1.143.2.1 2008/01/18 19:12:31 debug Exp $
- *
  *  CPU-related definitions.
  */
 
@@ -294,7 +292,7 @@ struct cpu_family {
 	/*  Print architecture-specific function call arguments.
 	    (This is called for each function call, if running with -t.)  */
 	void			(*functioncall_trace)(struct cpu *,
-				    uint64_t f, int n_args);
+				    int n_args);
 };
 
 
